from unittest.mock import call, patch

from arkindex.documents.tasks import reindex_corpus
from arkindex.project.tests import FixtureTestCase

EXPECTED_EMAIL_BODY = """
Hello Test user,

The search index build you started on project Unit Tests has been finished successfully.

Indexed elements can be explored from the search interface on Arkindex.

--
Arkindex

"""


class TestReindexCorpus(FixtureTestCase):

    def test_no_rq_job(self):
        with self.assertRaises(AssertionError) as ctx:
            reindex_corpus(corpus_id=self.corpus.id)
        self.assertEqual(str(ctx.exception), "This task can only be run in a RQ job context.")

    @patch("arkindex.documents.tasks.get_current_job")
    def test_no_user_id(self, job_mock):
        job_mock.return_value.user_id = None
        with self.assertRaises(AssertionError) as ctx:
            reindex_corpus(corpus_id=self.corpus.id)
        self.assertEqual(str(ctx.exception), "This task requires a user ID to be defined on the RQ job.")

    @patch("arkindex.documents.tasks.send_mail")
    @patch("arkindex.documents.tasks.get_current_job")
    @patch("arkindex.documents.tasks.Indexer")
    def test_run_drop_false(self, indexer_mock, job_mock, send_mail_mock):
        job_mock.return_value.user_id = self.user.id
        reindex_corpus(corpus_id=self.corpus.id, drop=False)

        self.assertEqual(indexer_mock.call_count, 1)
        self.assertEqual(indexer_mock.call_args, call(self.corpus.id))
        self.assertEqual(indexer_mock.return_value.drop_index.call_count, 0)
        self.assertEqual(indexer_mock.return_value.setup.call_count, 1)
        self.assertEqual(indexer_mock.return_value.index.call_count, 1)

        self.assertListEqual(send_mail_mock.call_args_list, [
            call(
                subject=f"Project {self.corpus.name} was successfully indexed",
                message=EXPECTED_EMAIL_BODY,
                from_email=None,
                recipient_list=[self.user.email],
                fail_silently=True,
            )
        ])

    @patch("arkindex.documents.tasks.send_mail")
    @patch("arkindex.documents.tasks.get_current_job")
    @patch("arkindex.documents.tasks.Indexer")
    def test_run(self, indexer_mock, job_mock, send_mail_mock):
        job_mock.return_value.user_id = self.user.id
        reindex_corpus(corpus_id=self.corpus.id)

        self.assertEqual(indexer_mock.call_count, 1)
        self.assertEqual(indexer_mock.call_args, call(self.corpus.id))
        self.assertEqual(indexer_mock.return_value.drop_index.call_count, 1)
        self.assertEqual(indexer_mock.return_value.setup.call_count, 1)
        self.assertEqual(indexer_mock.return_value.index.call_count, 1)

        self.assertListEqual(send_mail_mock.call_args_list, [
            call(
                subject=f"Project {self.corpus.name} was successfully indexed",
                message=EXPECTED_EMAIL_BODY,
                from_email=None,
                recipient_list=[self.user.email],
                fail_silently=True,
            )
        ])
