import tempfile
from pathlib import Path
from unittest.mock import MagicMock, PropertyMock, call, patch, seal

from django.test import override_settings

from arkindex.ponos.models import Farm, State, Task
from arkindex.ponos.tasks import run_docker_task
from arkindex.process.models import ProcessMode, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase


class TestModels(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.farm = Farm.objects.create(name="Farm")
        cls.process = cls.corpus.processes.create(
            creator=cls.user,
            mode=ProcessMode.Workers,
            corpus=cls.corpus,
            farm=cls.farm,
        )
        cls.worker_version1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_version2 = WorkerVersion.objects.get(worker__slug="dla")
        cls.run1 = cls.process.worker_runs.create(version=cls.worker_version1, parents=[])
        cls.run2 = cls.process.worker_runs.create(version=cls.worker_version2, parents=[cls.run1.id])

    @override_settings(PONOS_RQ_EXECUTION=True)
    @patch("arkindex.ponos.tasks.run_task_rq.delay")
    def test_run_process_schedules_tasks(self, run_task_mock):
        run_task_mock.side_effect = ["init_job", "job1", "job2"]
        # on_commit will not be called automatically as savepoints are used for tests transactions
        with self.captureOnCommitCallbacks(execute=True):
            self.process.run()
        init, t1, t2 = self.process.tasks.order_by("depth")
        self.assertEqual(run_task_mock.call_count, 3)
        call1, call2, call3 = run_task_mock.call_args_list
        self.assertEqual(call1, call(init))
        self.assertTupleEqual(call2.args, (t1,))
        self.assertListEqual(list(call2.kwargs.keys()), ["depends_on"])
        task1_depends = call2.kwargs["depends_on"]
        self.assertEqual(vars(task1_depends), {
            "dependencies": ["init_job"],
            "allow_failure": True,
            "enqueue_at_front": False,
        })
        self.assertTupleEqual(call3.args, (t2,))
        self.assertListEqual(list(call3.kwargs.keys()), ["depends_on"])
        task2_depends = call3.kwargs["depends_on"]
        self.assertEqual(vars(task2_depends), {
            "dependencies": ["job1"],
            "allow_failure": True,
            "enqueue_at_front": False,
        })

    @override_settings(PONOS_RQ_EXECUTION=True)
    @patch("arkindex.ponos.tasks.upload_logs")
    @patch("arkindex.ponos.tasks.sleep")
    def test_run_docker_task_stop(self, sleep_mock, upload_logs_mock):
        task = self.process.tasks.create(
            slug="something",
            image="image",
            env={"PATH": "/dev/null"},
            run=0,
            depth=0,
            state=State.Pending,
        )

        client_mock = MagicMock()
        container = client_mock.containers.run.return_value

        # The first two calls occur while the task is running and should not result in any upload.
        # Second call occurs after the container has been stopped.
        container.logs.side_effect = [b"", b"", b"^C"]

        # This will be accessed after the container has been stopped
        container.attrs = {
            "State": {
                # Stopping a container will likely cause a non-zero code, but it should be ignored
                # and the task should be marked as Stopped anyway
                "ExitCode": 1,
            }
        }

        def set_stopping(*args):
            # sleep() is called after one iteration of the running task loop is complete.
            # We use this mock to both not wait during tests and mark the task as stopping.
            task.refresh_from_db()
            self.assertEqual(task.state, State.Running)
            # We use .update() and not Task.save to ensure run_docker_task properly reloads the task from the DB.
            Task.objects.filter(id=task.id).update(state=State.Stopping)

        # Set up all the remaining attributes, then seal the mocks so nonexistent attributes can't be accessed
        sleep_mock.side_effect = set_stopping
        seal(sleep_mock)

        client_mock.images.pull.return_value = None
        container.reload.return_value = None
        container.stop.return_value = None
        # Sealing is not friendly with PropertyMocks, so we put a placeholder first
        container.status = None
        seal(client_mock)
        # Limit the amount of times container.status can be accessed, so we can't get stuck in an infinite loop
        type(container).status = PropertyMock(side_effect=[
            "running", # Loop that checks whether the container is `created` and needs to be awaited
            "running", # First iteration of the running task loop
            "running", # Second iteration where the task is Stopping
        ])

        upload_logs_mock.return_value = None
        seal(upload_logs_mock)

        with tempfile.TemporaryDirectory() as temp_dir:
            run_docker_task(client_mock, task, Path(temp_dir))

        task.refresh_from_db()
        self.assertEqual(task.state, State.Stopped)

        self.assertEqual(client_mock.images.pull.call_count, 1)
        self.assertEqual(client_mock.images.pull.call_args, call("image"))
        self.assertEqual(client_mock.containers.run.call_count, 1)
        self.assertEqual(client_mock.containers.run.call_args, call(
            "image",
            environment={"PATH": "/dev/null", "PONOS_DATA": "/data"},
            detach=True,
            network="host",
            volumes={temp_dir: {"bind": "/data", "mode": "rw"}},
        ))

        self.assertEqual(container.reload.call_count, 2)
        self.assertEqual(container.reload.call_args_list, [call(), call()])
        self.assertEqual(container.stop.call_count, 1)
        self.assertEqual(container.stop.call_args, call())
        self.assertEqual(container.logs.call_count, 3)
        self.assertEqual(container.logs.call_args_list, [call(), call(), call()])

        self.assertEqual(upload_logs_mock.call_count, 1)
        self.assertEqual(upload_logs_mock.call_args, call(task, b"^C"))

        self.assertEqual(sleep_mock.call_count, 1)
        self.assertEqual(sleep_mock.call_args, call(1))
