"""
Arkindex-specific Django system checks.

If you make any changes to those checks, you should consider updating the system
checks documentation in the Arkindex documentation:
<https://doc.arkindex.org/deployment/checks>
"""
import os
import re
import sys

from django.core.checks import Critical, Error, Warning, register

from arkindex.process.models import ArkindexFeature, WorkerVersion


def only_runserver(func):
    """Decorator to run a system check only when running a dev server or deploying"""
    from django.conf import settings

    def wrapper(*args, **kwargs):
        if settings.ALL_CHECKS:
            return func(*args, **kwargs)
        return []

    return wrapper


ROUTE_SLASH_REGEX = re.compile(r"(?:<path:[^>]+>|/)$")


@register()
def api_urls_check(*args, **kwargs):
    """
    Check that all API URL patterns end with a mandatory slash,
    except those that end with a <path:…> parameter,
    since the ending slash would become part of the parameter.
    """
    try:
        from arkindex.project.api_v1 import api
    except ImportError as e:
        return [
            Error(
                f"Cannot import API URLs: {e}",
                id="arkindex.E001",
            )
        ]

    return [
        Error(
            "This API URL is missing a mandatory ending /",
            hint=str(url),
            id="arkindex.E001",
        )
        for url in api
        if not ROUTE_SLASH_REGEX.search(str(url.pattern))
    ]


@register()
@only_runserver
def local_imageserver_check(*args, **kwargs):
    """
    Check that the configured local ImageServer exists
    """
    from django.conf import settings

    from arkindex.images.models import ImageServer
    local_id = settings.LOCAL_IMAGESERVER_ID

    try:
        ImageServer.objects.get(id=local_id)
    except ImageServer.DoesNotExist:
        return [Warning(
            f"Local ImageServer with ID {local_id} does not exist.",
            hint=f"settings.LOCAL_IMAGESERVER_ID = {local_id}",
            id="arkindex.W009",
        )]

    return []


@register()
def ponos_env_check(*args, **kwargs):
    """
    Check that the base Ponos env seems legit
    """
    from django.conf import settings
    errors = []

    env = settings.PONOS_DEFAULT_ENV.copy()
    for variable in ("ARKINDEX_API_URL", "ARKINDEX_API_CSRF_COOKIE"):
        if variable not in env:
            errors.append(Warning(
                f"The {variable} environment variable should be defined "
                "to allow API client autoconfiguration in Ponos tasks.",
                hint=f"`ponos.default_env` in {settings.CONFIG_PATH}",
                id="arkindex.W006",
            ))

    if "ARKINDEX_API_TOKEN" in env:
        errors.append(Warning(
            "Defining a default value for the ARKINDEX_API_TOKEN variable may introduce a security issue "
            "by allowing Ponos tasks to bypass the Ponos task authentication. Consider removing it.",
            hint=f"`ponos.default_env.ARKINDEX_API_TOKEN` in {settings.CONFIG_PATH}",
            id="arkindex.W013",
        ))

    return errors


@register()
@only_runserver
def s3_check(*args, **kwargs):
    from django.conf import settings
    aws_settings = {
        "AWS_ACCESS_KEY": "AWS access key ID",
        "AWS_SECRET_KEY": "AWS secret key",
        "AWS_THUMBNAIL_BUCKET": "S3 thumbnails bucket name",
        "AWS_STAGING_BUCKET": "S3 staging bucket name",
        "AWS_EXPORT_BUCKET": "S3 export bucket name",
    }
    errors = []
    for name, display_name in aws_settings.items():
        value = getattr(settings, name, None)
        if not value:
            errors.append(Error(
                f"{display_name} is missing; all S3-related features will fail.",
                hint=f"settings.{name} = {value!r}",
                id="arkindex.E011",
            ))

    return errors


@register()
@only_runserver
def public_hostname_check(*args, **kwargs):
    from django.conf import settings
    if settings.PUBLIC_HOSTNAME is None:
        return [Warning(
            "The public_hostname setting is missing; absolute URLs may not be correctly generated.",
            hint="settings.PUBLIC_HOSTNAME",
            id="arkindex.W008",
        )]
    return []


@register()
def python_version_check(*args, **kwargs):
    if sys.version_info.minor < 10:
        return [
            Critical(
                "Arkindex requires Python 3.10 or later.",
                id="arkindex.C001",
            )
        ]
    return []


@register()
@only_runserver
def ingest_check(*args, **kwargs):
    from django.conf import settings
    warnings = []

    # Do not run any check when the S3 ingest feature flag is disabled
    if not settings.ARKINDEX_FEATURES["ingest"]:
        return []

    if settings.INGEST_S3_ACCESS_KEY is None:
        warnings.append(Warning(
            "INGEST_S3_ACCESS_KEY is not set. The S3 ingest feature will not work.",
            id="arkindex.W010",
        ))

    if settings.INGEST_S3_SECRET_KEY is None:
        warnings.append(Warning(
            "INGEST_S3_SECRET_KEY is not set. The S3 ingest feature will not work.",
            id="arkindex.W010",
        ))

    if settings.INGEST_S3_REGION is None and settings.INGEST_S3_ENDPOINT is None:
        warnings.append(Warning(
            "Neither INGEST_S3_REGION nor INGEST_S3_ENDPOINT are set. The S3 ingest feature will not work.",
            id="arkindex.W010",
        ))

    if settings.INGEST_IMAGESERVER_ID is None:
        warnings.append(Warning(
            "INGEST_IMAGESERVER_ID is not set. The S3 ingest feature will not work.",
            id="arkindex.W010",
        ))
    else:
        from arkindex.images.models import ImageServer
        try:
            ImageServer.objects.get(id=settings.INGEST_IMAGESERVER_ID)
        except ImageServer.DoesNotExist:
            warnings.append(Warning(
                f"ImageServer with ID {settings.INGEST_IMAGESERVER_ID} does not exist. The S3 ingest feature will not work.",
                hint=f"settings.INGEST_IMAGESERVER_ID = {settings.INGEST_IMAGESERVER_ID}",
                id="arkindex.W010",
            ))

    return warnings


@register()
def botocore_config_check(*args, **kwargs):
    """
    When not all S3 settings have both an access key ID, secret access key
    and either an endpoint or a region set, botocore might use some environment variables
    or configuration files to fill in the blanks. This can cause confusion for developers
    as some unit tests might fail or some other behaviors might be different, and can cause
    data leaks in production where some credentials might be used unexpectedly.
    """
    from django.conf import settings

    required_settings = [
        settings.AWS_ACCESS_KEY,
        settings.AWS_SECRET_KEY,
        settings.AWS_ENDPOINT or settings.AWS_REGION,
        settings.INGEST_S3_ACCESS_KEY,
        settings.INGEST_S3_SECRET_KEY,
        settings.INGEST_S3_ENDPOINT or settings.INGEST_S3_REGION,
    ]
    if all(setting is not None for setting in required_settings):
        # All settings are set, no complaints
        return []

    forbidden_variables = {
        "AWS_ACCESS_KEY_ID",
        "AWS_CONFIG_FILE",
        "AWS_CREDENTIAL_FILE",
        "AWS_DEFAULT_PROFILE",
        "AWS_DEFAULT_REGION",
        "AWS_EXECUTION_ENV",
        "AWS_PROFILE",
        "AWS_REGION",
        "AWS_SECRET_ACCESS_KEY",
        "AWS_SECURITY_TOKEN",
        "AWS_SESSION_TOKEN",
        "BOTO_CONFIG",
    }
    forbidden_files = {
        "~/.aws/config",
        "~/.aws/credentials",
    }
    warnings = []

    found_variables = forbidden_variables & set(os.environ.keys())
    warnings.extend([
        Warning(
            f"The {variable} environment variable is set, and not all S3 settings are set.\n"
            "This variable might override the Arkindex settings and cause unexpected behavior.",
            id="arkindex.W011",
        )
        for variable in found_variables
    ])

    for path in forbidden_files:
        if os.path.exists(os.path.expanduser(path)):
            warnings.extend([
                Warning(
                    f"The {path} file exists, and not all S3 settings are set.\n"
                    "The settings in this file might override the Arkindex settings and cause unexpected behavior.",
                    id="arkindex.W011",
                )
            ])

    return warnings


@register()
@only_runserver
def signup_default_group_check(*args, **kwargs):
    """
    Check that, if it is set, SIGNUP_DEFAULT_GROUP is a group that exists
    """
    from django.conf import settings

    from arkindex.users.models import Group
    group_id = settings.SIGNUP_DEFAULT_GROUP

    if not group_id:
        return []

    try:
        Group.objects.get(id=group_id)
    except Group.DoesNotExist:
        return [Warning(
            f"Group object with ID {group_id} does not exist",
            hint=f"settings.SIGNUP_DEFAULT_GROUP = {group_id}",
            id="arkindex.W012",
        )]

    return []

@register()
@only_runserver
def init_elements_check(*args, **kwargs):
    try:
        WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
    except WorkerVersion.DoesNotExist:
        return [Warning(
            "There is no worker version available to initialize elements.\n"
            "This worker version is required to initialize elements when running a process on Arkindex.\n"
            "Starting or retrying processes will fail.",
            hint="Run `arkindex update_system_workers` or set a worker version's `feature` to `init_elements`.",
            id="arkindex.W014",
        )]

    return []

@register()
@only_runserver
def ingest_version_check(*args, **kwargs):
    from django.conf import settings

    # Do not run any check when the S3 ingest feature flag is disabled
    if not settings.ARKINDEX_FEATURES["ingest"]:
        return []

    try:
        WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest)
    except WorkerVersion.DoesNotExist:
        return [Warning(
            "There is no worker version available to import elements from S3.\n"
            "Starting or retrying S3 import processes will fail.",
            hint="Run `arkindex update_system_workers` or set a worker version's `feature` to `s3_ingest`.",
            id="arkindex.W015",
        )]

    return []
